from src.preprocessor import Preprocessor
import pandas as pd


def test_preprocessor_with_path():
    instance = Preprocessor(path="data/ml-1m/users.dat", delimiter="::")
    assert type(instance.data) == pd.DataFrame


def test_preprocessor_with_df(user_df):
    instance = Preprocessor(df=user_df)
    assert type(instance.data) == pd.DataFrame


def test_onehot_encode(user_preprocessor):
    user_preprocessor.onehot_encode(categorical_columns=["gender"])

    # grab a reference to the data for easy access
    data = user_preprocessor.data

    assert "ftr_M" in list(data.columns)
    assert "ftr_F" in list(data.columns)
    assert "ftr_U" in list(data.columns)
    assert data["ftr_M"].sum() == 2
    assert data["ftr_F"].sum() == 2
    assert data["ftr_U"].sum() == 1


def test_normalize_linear(user_preprocessor):
    user_preprocessor.normalize(["age"], methods=["linear"])

    # grab a reference to the data for easy access
    data = user_preprocessor.data

    assert (1 >= data["age"]).all()
    assert (0 <= data["age"]).all()


def test_normalize_z_score(user_preprocessor):
    user_preprocessor.normalize(["age"], methods=["z-score"])

    # grab a reference to the data for easy access
    data = user_preprocessor.data

    assert (1.8 >= data["age"]).all()
    assert (-1.5 <= data["age"]).all()


def test_handle_missing_values_by_dropping(item_preprocessor):
    # grab a reference to the data for easy access
    data = item_preprocessor.data

    item_preprocessor.handle_missing_values(strategy="drop")

    assert len(data) == 4


def test_handle_missing_values_with_mean(item_preprocessor):
    data = item_preprocessor.data

    category_mode = data["category"].mode()[0]

    price_mean = data["price"].mean()
    item_preprocessor.handle_missing_values(strategy="mean")

    assert data.loc[6, "price"] == price_mean
    assert data.loc[4, "category"] == category_mode


def test_multilabel_binarize(item_preprocessor):
    item_preprocessor.multilabel_binarize(["tags"])
    data = item_preprocessor.data

    assert "ftr_T1" in data.columns
    assert "ftr_T2" in data.columns
    assert "ftr_T3" in data.columns
    assert "ftr_T4" in data.columns
    assert "ftr_T5" in data.columns



import click
import pandas as pd

from .algo.content_based.knn import KNN


@click.group()
def main():
    pass


@main.command(
    help="Builds a recommendation model using the given algorithm on the given dataset"
)
@click.argument(
    "algorithm",
    type=click.Choice(
        ["content-knn", "user-cf-knn", "item-cf-knn"],
    ),
)
@click.argument("dataset", type=click.Path(exists=True))
@click.option(
    "--model-path",
    default="model.pkl",
    help="Path to the file where the model will be serialized",
    type=click.Path(),
)
def build(algorithm, dataset, model_path):
    click.echo("Building recommender system...")
    click.echo("Algorithm: %s" % algorithm)
    click.echo("Dataset: %s" % dataset)
    click.echo("Model path : %s" % model_path)

    df = pd.read_csv(dataset)

    if algorithm == "content-knn":
        model = KNN(df)
        model.serialize(model_path)
        print("Finished building content based KNN model.")


if __name__ == "__main__":
    main()



from setuptools import find_packages, setup

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="rstk",
    version="0.1.3",
    package_dir={"rstk": "src"},
    install_requires=[
        "click",
        "pandas",
        "numpy",
        "scikit-learn",
    ],
    entry_points="""
        [console_scripts]
        rstk=rstk.cli:main
    """,
    author="Nikola Ratinac",
    author_email="nikola.ratinac@gmail.com",
    description="Lightweight recommender system toolkit",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="3-clause BSD",
    url="https://github.com/rat-nick/rstk",
)



from typing import List, Literal

import pandas as pd

z_score = lambda x, mean, std: (x - mean) / std
linear = lambda x, min, max: (x - min) / (max - min)


class Preprocessor:
    def __init__(self, path: str = None, df: pd.DataFrame = None, delimiter: str = ","):
        if path != None:
            self.data = self.load(path, delimiter)
        elif df is not None:
            self.data = df
        else:
            raise ValueError("Must provide either path to data or dataframe object")

    def load(self, path: str, delimiter: str = ",") -> pd.DataFrame:
        return pd.read_csv(path, delimiter=delimiter, header=0, engine="python")

    def preprocess(
        self,
        missing_value_strategy: Literal["drop", "mean", "median", "mode"] = "drop",
        normalization_columns: List[str] = None,
        normalization_methods: List[Literal["linear", "z-score"]] = None,
        categorical_columns: List[str] = None,
        multilabel_columns: List[str] = None,
        feature_columns: List[str] = None,
    ) -> pd.DataFrame:
        self.handle_missing_values(missing_value_strategy)

        if normalization_columns != None:
            self.normalize(normalization_columns, normalization_methods)

        if categorical_columns != None:
            self.onehot_encode(categorical_columns)

        if multilabel_columns != None:
            self.multilabel_binarize(multilabel_columns)
        # get all the columns that were generated via feature engineering
        res_df = self.data.filter(regex="^ftr_*")

        return res_df.join(self.data[[feature_columns]])

    def handle_missing_values(
        self, strategy: Literal["drop", "mean", "median", "mode"] = "drop"
    ):
        if strategy == "drop":
            self.data.dropna(inplace=True)
        else:
            self.fillna(strategy=strategy)

    def fillna(self, strategy: Literal["mean", "median", "mode"] = "drop"):
        for col in self.data.columns:
            fill_val = self.determine_fill_value(col, strategy)
            self.data[col] = self.data[col].fillna(fill_val)

    def determine_fill_value(self, col, strategy):
        dtype = self.data[col].dtype

        fill_val = 0
        if dtype == "object":
            fill_val = self.data[col].mode()[0]
        if dtype in ["int64", "float64"]:
            if strategy == "mean":
                fill_val = self.data[col].mean()
            if strategy == "median":
                fill_val = self.data[col].median()
            if strategy == "mode":
                fill_val == self.data[col].mode()

        return fill_val

    def normalize(
        self,
        normalization_columns: List[str] = [],
        methods: List[Literal["linear", "z-score"]] = None,
    ):
        """Normalizes the given columns of the DataFrame using the specified method."""

        if len(normalization_columns) > 0 and methods == None:
            methods = len(normalization_columns) * ["linear"]
        if len(methods) != len(normalization_columns):
            raise ValueError("Method and column lists must be of equal length.")

        for col, meth in zip(normalization_columns, methods):
            self.normalize_column(col, meth)

    def normalize_column(self, col, meth):
        if meth == "z-score":
            # perform z-score normalization
            self.data[col] = z_score(
                self.data[col], self.data[col].mean(), self.data[col].std()
            )

        elif meth == "linear":
            # perform linear normalization
            self.data[col] = linear(
                self.data[col], self.data[col].min(), self.data[col].max()
            )

    def onehot_encode(self, categorical_columns: List[str] = []):
        self.data = pd.get_dummies(self.data, columns=categorical_columns, prefix="ftr")

    def multilabel_binarize(self, multilabel_columns: List[str] = [], sep="|"):
        for col in multilabel_columns:
            binarized_labels = self.data[col].str.get_dummies(sep).add_prefix("ftr_")
            self.data.drop(columns=col, inplace=True)
            self.data = pd.concat([self.data, binarized_labels], axis=1)
        print(self.data)

    def select_features(self, columns: List[str] = []) -> pd.DataFrame:
        pass



import os

import numpy as np
import pandas as pd
import pytest

from ..src.algo.content_based.knn import KNN


@pytest.fixture
def feature_data():
    # generate 10 rows and 5 columns of binary data
    data = {
        "id": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12],
        "ftr_1": [0, 1, 0, 1, 1, 0, 0, 1, 0, 1, 0, 0],
        "ftr_2": [0, 0, 0, 1, 0, 1, 1, 0, 1, 0, 0, 0],
        "ftr_3": [0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0],
        "ftr_4": [1, 0, 1, 0, 0, 1, 0, 0, 1, 0, 0, 0],
        "ftr_5": [1, 0, 1, 0, 0, 0, 1, 0, 0, 1, 0, 1],
    }
    df = pd.DataFrame(data)
    return df


def test_knn_constructors(feature_data):
    knn = KNN(data=feature_data, id_column="id")

    assert knn.features.shape == (12, 5)

    knn = KNN(
        data=feature_data,
        id_column="id",
        feature_columns=["ftr_1", "ftr_3"],
    )

    assert knn.features.shape == (12, 2)

    knn = KNN(data=feature_data)

    assert knn.features.shape == (12, 5)

    with pytest.raises(ValueError):
        knn = KNN(
            data=feature_data,
            id_column="id",
            feature_prefix="nonexistent",
        )


@pytest.fixture
def knn(feature_data):
    return KNN(data=feature_data, id_column="id")


def test_get_recommendations(knn):
    user_profile = np.array([1, 1, 0, 1, 0])
    recs = knn.get_recommendations(user_profile, 3)
    assert recs == [9, 6, 4]


def test_get_recommendations_with_list_user_profile(knn):
    user_profile = [1, 1, 0, 1, 0]
    recs = knn.get_recommendations(user_profile, 3)
    assert recs == [9, 6, 4]


def test_get_similar_items_with_incorrect_user_profile(knn):
    user_profile = ["h", "e", "l", "l", "o"]
    with pytest.raises(ValueError):
        knn.get_recommendations(user_profile, 3)


def test_model_searialization(knn):
    path = "./example.pkl"
    knn.serialize(path)

    assert os.path.exists(path)

    deserialized_knn = KNN.deserialize(path=path)

    assert type(deserialized_knn) == KNN

    os.remove(path)


def test_model_serialization_with_nested_path(knn):
    path = "nested/path.pkl"
    knn.serialize(path)

    assert os.path.exists(path)

    deserialized_knn = KNN.deserialize(path=path)

    assert type(deserialized_knn) == KNN

    os.remove(path)
    os.removedirs(os.path.dirname(path))



import os
import pickle
from typing import List

import numpy as np
import pandas as pd
from sklearn.metrics.pairwise import cosine_similarity


class KNN:
    """Content based item recommendation system using the k-nearest neighbors algorithm"""

    def __init__(
        self,
        data: pd.DataFrame = None,
        id_column: str = None,
        feature_columns: List[str] = None,
        feature_prefix: str = "ftr",
    ):
        """
        Initialize the KNN class.

        Parameters
        ----------
        data : pd.DataFrame
            The data frame containing the items and their features.
        id_column : str
            The name of the column in the data frame that contains the item IDs.
        feature_prefix : str, optional
            The prefix of the columns in the data frame that contain the features, by default "ftr".
        feature_columns : List[str], optional
            A list of the names of the columns in the data frame that contain the features, by default None. If None, all columns with the specified prefix are used.

        Raises
        ------
        ValueError
            If no features are specified and no columns with the specified prefix are found.
        """
        self.data = data

        if id_column is not None:
            self.data = self.data.set_index(id_column)

        self.translate_ids()

        self.init_features(feature_columns, feature_prefix)

    def init_features(self, feature_columns, feature_prefix):
        """
        Initializes the features based on the provided feature columns and prefix.

        Parameters:
            feature_columns (list): List of feature columns to initialize the features.
            feature_prefix (str): Prefix to filter the feature columns.

        Raises:
            ValueError: If no features are selected.

        Returns:
            None
        """
        if feature_columns is None:
            self.features = self.data.filter(regex=f"^{feature_prefix}*")
        else:
            self.features = self.data[feature_columns]

        if len(self.features.columns) == 0:
            raise ValueError("No features selected")

        # translate the features to a numpy array
        self.features = self.features.to_numpy()

    def translate_ids(self):
        """
        Translates the raw item IDs into unique inner IDs and vice versa.

        The inner IDs are used to represent the items in the recommendation model, while the raw IDs
        are used to retrieve the original data from the database.

        Parameters
        ----------
        None

        Returns
        -------
        None

        """
        # dict where the key is the raw item id and the value is a unique integer for each item
        self.raw2innerID = {}
        # dict where the key is the unique integer for each item and the value is the raw item id
        self.inner2rawID = {}
        for id, _ in self.data.iterrows():
            if id not in self.raw2innerID.keys():
                innerID = len(self.raw2innerID)
                self.raw2innerID[id] = innerID
                self.inner2rawID[innerID] = id

    def get_recommendations(
        self,
        user_profile: np.array,
        k: int = 10,
    ) -> List[any]:
        """
        Get the k most similar items to the given user profile using cosine similarity.

        Args:
            user_profile (np.array): The user profile for which to find similar items.
            k (int): The number of similar items to return. Defaults to 10.

        Returns:
            List[any]: A list of the k most similar items to the user profile.
        """
        if user_profile is None:
            raise ValueError("Must provide the user profile")

        if type(user_profile) is not np.array:
            user_profile = np.array(user_profile)

        # calculate the similarity of all items to the user profile
        similarity = cosine_similarity(user_profile.reshape(1, -1), self.features)

        # get the indicies of the k most similar items
        indicies = np.argsort(similarity, axis=1)[0][::-1]

        # translate the indicies into raw IDs
        return [self.inner2rawID[i] for i in indicies][:k]

    def serialize(self, path: str):
        """
        Serialize the model to a file.

        Args:
            path (str): The file path to serialize the object to.

        Returns:
            None
        """

        # check if its the current directory
        if "/" not in path:
            path = "./" + path
        # check if the path exists and if not create it
        if not os.path.exists(path):
            os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, "wb") as f:
            pickle.dump(self, f)

    @staticmethod
    def deserialize(path: str) -> "KNN":
        """
        Deserialize a KNN model from the given path using pickle.

        Args:
            path (str): The path to the serialized KNN model.

        Returns:
            KNN: The deserialized KNN model.
        """
        with open(path, "rb") as f:
            return pickle.load(f)
